@@ -16,7 +16,6 @@ You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 -}
-{-# LANGUAGE CPP #-}
 module Bustle.Diagram
   ( Diagram
 
@@ -37,32 +36,35 @@ module Bustle.Diagram
   , columnWidth
   , timestampAndMemberWidth
   , firstColumnOffset
+  , eventHeight
 
   -- Displaying diagrams
   , diagramDimensions
   , topLeftJustifyDiagram
+  , translateDiagram
   , drawDiagram
   , drawRegion
   )
 where
 
-import Data.Maybe (maybe)
 import Data.List (unzip4)
 import Control.Arrow ((&&&))
 import Control.Applicative ((<$>), (<*>))
-import Control.Monad (forM_)
 
 import Control.Monad.Reader
 
 import Graphics.Rendering.Cairo
 import Graphics.UI.Gtk.Cairo (cairoCreateContext, showLayout)
-#if MIN_VERSION_gtk(0,11,0)
 import Graphics.Rendering.Pango.Layout
 import Graphics.Rendering.Pango.Font
-#else
-import Graphics.UI.Gtk.Pango.Layout
-import Graphics.UI.Gtk.Pango.Font
-#endif
+
+import qualified Bustle.Markup as Markup
+import Bustle.Markup (Markup)
+import Bustle.Util
+import Bustle.Types (ObjectPath, InterfaceName, MemberName)
+
+-- Sorry Mum
+import System.IO.Unsafe (unsafePerformIO)
 
 type Point = (Double, Double)
 type Rect = (Double, Double, Double, Double)
@@ -92,7 +94,10 @@ data Colour = Colour Double Double Double
 data Shape = Header { strs :: [String]
                     , shapex, shapey :: Double
                     }
-           | MemberLabel { labelPath, labelMember :: String
+           | MemberLabel { labelPath :: ObjectPath
+                         , labelInterface :: Maybe InterfaceName
+                         , labelMember :: MemberName
+                         , shapeIsReturn :: Bool
                          , shapex :: Double -- The coordinates of the *centre*
                          , shapey :: Double -- of the label
                          }
@@ -100,23 +105,33 @@ data Shape = Header { strs :: [String]
                             , shapex :: Double -- The coordinates of the
                             , shapey :: Double -- *centre* of the timestamp
                             }
-           | ClientLine { shapex, shapey1, shapey2 :: Double }
+           | ClientLines { shapexs :: NonEmpty Double -- The x-coordinates of the lines to draw
+                         , shapey1, shapey2 :: Double
+                         }
            | Rule { shapex1, shapex2, shapey :: Double }
            | Arrow { shapecolour :: Maybe Colour
                    , arrowhead :: Arrowhead
                    , shapex1, shapex2, shapey :: Double
                    }
            | SignalArrow { shapex1, epicentre, shapex2, shapey :: Double }
+           | DirectedSignalArrow { epicentre, shapex, shapey :: Double }
            | Arc { topx, topy, bottomx, bottomy :: Double
                  , arcside :: Side
                  , caption :: String
                  }
-  deriving (Show, Read, Eq)
+           | Highlight { highlightRegion :: Rect
+                       }
+  deriving (Show, Eq)
 
 -- Smart constructors for TimestampLabel and MemberLabel that fill in the
 -- hardcoded (spit) x coordinates.
-memberLabel :: String -> String -> Double -> Shape
-memberLabel p m y = MemberLabel p m memberx y
+memberLabel :: ObjectPath
+            -> Maybe InterfaceName
+            -> MemberName
+            -> Bool   -- ^ True if this is a return; False if it's a call
+            -> Double -- ^ y-coordinate
+            -> Shape
+memberLabel p i m isReturn y = MemberLabel p i m isReturn memberx y
 
 timestampLabel :: String -> Double -> Shape
 timestampLabel s y = TimestampLabel s timestampx y
@@ -146,13 +161,14 @@ mapX f s = case s of
     Arc {}         -> s { topx = f (topx s)
                         , bottomx = f (bottomx s)
                         }
+    ClientLines {} -> s { shapexs = mapNonEmpty f (shapexs s) }
     _              -> s { shapex = f (shapex s) }
 
 mapY f s = case s of
     Arc {}        -> s { topy = f (topy s)
                        , bottomy = f (bottomy s)
                        }
-    ClientLine {} -> s { shapey1 = f (shapey1 s)
+    ClientLines {} -> s { shapey1 = f (shapey1 s)
                        , shapey2 = f (shapey2 s)
                        }
     _             -> s { shapey = f (shapey s) }
@@ -203,7 +219,9 @@ headerHeight = fromIntegral . (10 *) . length
 
 bounds :: Shape -> Rect
 bounds s = case s of
-  ClientLine {} -> (shapex s, shapey1 s, shapex s, shapey2 s)
+  ClientLines {} ->
+    let xs = nonEmptyToList (shapexs s)
+    in  (minimum xs, shapey1 s, maximum xs, shapey2 s)
   Rule {} -> (shapex1 s, shapey s, shapex2 s, shapey s)
   Arrow {} ->
     let (x1, x2) = xMinMax s
@@ -214,17 +232,22 @@ bounds s = case s of
     let (x1, x2) = xMinMax s
         (y1, y2) = (subtract 5) &&& (+5) $ shapey s
     in (x1, y1, x2, y2)
+  DirectedSignalArrow {} ->
+    let (x1, x2) = minMax (epicentre s, shapex s)
+        (y1, y2) = (subtract 5) &&& (+5) $ shapey s
+    in (x1, y1, x2, y2)
   Arc { topx=x1, bottomx=x2, topy=y1, bottomy=y2 } ->
     let ((cx, _), (dx, _)) = arcControlPoints s
        -- FIXME: magic 5 makes the bounding box include the text
     in (min x1 cx, y1, max x2 dx, y2 + 5)
   TimestampLabel { shapex=x, shapey=y } -> fromCentre x y timestampWidth
-  MemberLabel _ _ x y -> fromCentre x y memberWidth
+  MemberLabel { shapex=x, shapey=y } -> fromCentre x y memberWidth
   Header { strs = ss, shapex = x, shapey = y} ->
     let width = columnWidth
         height = headerHeight ss
     in (x - width / 2, y,
         x + width / 2, y + height)
+  Highlight r -> r
 
 intersects :: Rect -> Rect -> Bool
 intersects (x,y,w,z) (x', y', w', z') =
@@ -258,17 +281,20 @@ diagramDimensions shapes = (x2 - x1, y2 - y1)
   where
     ((x1, y1), (x2, y2)) = diagramBounds shapes
 
-topLeftJustifyDiagram :: Diagram -- ^ the original diagram
-                      -> (Double, Diagram) -- ^ the diagram transformed to be
-                                           --   in positive space, and the
-                                           --   x-axis shift necessary to do so
-topLeftJustifyDiagram shapes = (negate x1, shapes')
+topLeftJustifyDiagram
+    :: Diagram -- ^ the original diagram
+    -> ((Double, Double), Diagram) -- ^ the diagram transformed to be in
+                                   --   positive space, and the (x, y)-axis
+                                   --   shifts necessary to do so
+topLeftJustifyDiagram shapes =
+    (translation, shapes')
   where
-    shapes'       = transformDiagram (negate x1, negate y1) shapes
     ((x1, y1), _) = diagramBounds shapes
+    translation   = (negate x1, negate y1)
+    shapes'       = translateDiagram translation shapes
 
-transformDiagram :: (Double, Double) -> (Diagram -> Diagram)
-transformDiagram (x, y) = map (mapX (+ x) . mapY (+ y))
+translateDiagram :: (Double, Double) -> (Diagram -> Diagram)
+translateDiagram (x, y) = map (mapX (+ x) . mapY (+ y))
 
 drawDiagramInternal :: (Shape -> Bool) -- ^ A filter for the shapes
                     -> Bool -- ^ True to draw canvas items' bounding boxes
@@ -314,23 +340,30 @@ draw s = draw' s
           Arc {} -> let ((cx, cy), (dx, dy)) = arcControlPoints s
                     in drawArc cx cy dx dy <$>
                           topx <*> topy <*> bottomx <*> bottomy <*> caption
-          SignalArrow {} -> drawSignalArrow <$> epicentre <*> shapex1 <*>
-                              shapex2 <*> shapey
+          SignalArrow {} -> drawSignalArrow <$> epicentre
+                                            <*> Just . shapex1
+                                            <*> Just . shapex2
+                                            <*> shapey
+          DirectedSignalArrow { } -> drawDirectedSignalArrow <$> epicentre
+                                                             <*> shapex
+                                                             <*> shapey
           Arrow {} -> drawArrow <$> shapecolour <*> arrowhead <*> shapex1 <*>
                         shapex2 <*> shapey
           Header {} -> drawHeader <$> strs <*> shapex <*> shapey
           MemberLabel {} -> drawMember <$> labelPath
+                                       <*> labelInterface
                                        <*> labelMember
+                                       <*> shapeIsReturn
                                        <*> shapex
                                        <*> shapey
           TimestampLabel {} -> drawTimestamp <$> str
                                              <*> shapex
                                              <*> shapey
-          ClientLine {} -> drawClientLine <$> shapex <*> shapey1 <*> shapey2
+          ClientLines {} -> drawClientLines <$> shapexs <*> shapey1 <*> shapey2
           Rule {} -> drawRule <$> shapex1
                               <*> shapex2
                               <*> shapey
-
+          Highlight {} -> drawHighlight <$> highlightRegion
 
 halfArrowHead :: Arrowhead -> Bool -> Render ()
 halfArrowHead a left = do
@@ -354,21 +387,35 @@ drawArrow c a from to y = saved $ do
     halfArrowHead a (from < to)
     stroke
 
-drawSignalArrow :: Double -> Double -> Double -> Double -> Render ()
-drawSignalArrow e left right y = do
+drawDirectedSignalArrow :: Double -- ^ the signal emission source
+                        -> Double -- ^ signal target coordinate
+                        -> Double -- ^ vertical coordinate
+                        -> Render ()
+drawDirectedSignalArrow e x y
+    | x < e     = drawSignalArrow e (Just x) Nothing y
+    | otherwise = drawSignalArrow e Nothing (Just x) y
+
+drawSignalArrow :: Double -- ^ the signal emission source
+                -> Maybe Double -- ^ left-pointing arrow coordinate
+                -> Maybe Double -- ^ right-pointing arrow coordinate
+                -> Double -- ^ vertical coordinate
+                -> Render ()
+drawSignalArrow e mleft mright y = do
     newPath
     arc e y 5 0 (2 * pi)
     stroke
 
-    moveTo left y
-    arrowHead False
-    lineTo (e - 5) y
-    stroke
+    maybeM mleft $ \left -> do
+        moveTo left y
+        arrowHead False
+        lineTo (e - 5) y
+        stroke
 
-    moveTo (e + 5) y
-    lineTo right y
-    arrowHead True
-    stroke
+    maybeM mright $ \right -> do
+        moveTo (e + 5) y
+        lineTo right y
+        arrowHead True
+        stroke
 
 drawArc :: Double -> Double -> Double -> Double
         -> Double -> Double -> Double -> Double
@@ -383,27 +430,28 @@ drawArc cx cy dx dy x1 y1 x2 y2 cap = saved $ do
     stroke
 
     setSourceRGB 0 0 0
-    l <- mkLayout cap EllipsizeNone AlignLeft
+    l <- mkLayout (Markup.escape cap) EllipsizeNone AlignLeft
     (PangoRectangle _ _ textWidth _, _) <- liftIO $ layoutGetExtents l
     let tx = min x2 dx + abs (x2 - dx) / 2
     moveTo (if x1 > cx then tx - textWidth else tx) (y2 - 5)
     showLayout l
 
-font :: IO FontDescription
-font = do
+font :: FontDescription
+font = unsafePerformIO $ do
     fd <- fontDescriptionNew
     fontDescriptionSetSize fd 7
     fontDescriptionSetFamily fd "Sans"
     return fd
+{-# NOINLINE font #-}
 
 mkLayout :: (MonadIO m)
-         => String -> EllipsizeMode -> LayoutAlignment
+         => Markup -> EllipsizeMode -> LayoutAlignment
          -> m PangoLayout
 mkLayout s e a = liftIO $ do
     ctx <- cairoCreateContext Nothing
     layout <- layoutEmpty ctx
-    layoutSetMarkup layout s
-    layoutSetFontDescription layout . Just =<< font
+    layoutSetMarkup layout (Markup.unMarkup s)
+    layoutSetFontDescription layout (Just font)
     layoutSetEllipsize layout e
     layoutSetAlignment layout a
     return layout
@@ -416,30 +464,43 @@ withWidth m w = do
 
 drawHeader :: [String] -> Double -> Double -> Render ()
 drawHeader names x y = forM_ (zip [0..] names) $ \(i, name) -> do
-    l <- mkLayout name EllipsizeEnd AlignCenter `withWidth` columnWidth
+    l <- mkLayout (Markup.escape name) EllipsizeEnd AlignCenter `withWidth` columnWidth
     moveTo (x - (columnWidth / 2)) (y + i * h)
     showLayout l
   where h = 10
 
-drawMember :: String -> String -> Double -> Double -> Render ()
-drawMember s1 s2 x y = dm s1 (y - 10) >> dm s2 y
+drawMember :: ObjectPath
+           -> Maybe InterfaceName
+           -> MemberName
+           -> Bool
+           -> Double
+           -> Double
+           -> Render ()
+drawMember p i m isReturn x y = do
+    drawOne path (y - 10)
+    drawOne fullMethod y
   where
-    dm s y' = do
-      l <- mkLayout s EllipsizeStart AlignLeft `withWidth` memberWidth
+    drawOne markup y' = do
+      l <- mkLayout markup EllipsizeStart AlignLeft `withWidth` memberWidth
       moveTo (x - memberWidth / 2) y'
       showLayout l
 
+    path = (if isReturn then id else Markup.b) $ Markup.escape p
+    fullMethod =
+        (if isReturn then Markup.i else id) $ Markup.formatMember i m
+
 drawTimestamp :: String -> Double -> Double -> Render ()
 drawTimestamp ts x y = do
     moveTo (x - timestampWidth / 2) (y - 10)
-    showLayout =<< mkLayout ts EllipsizeNone AlignLeft `withWidth` timestampWidth
+    showLayout =<< mkLayout (Markup.escape ts) EllipsizeNone AlignLeft `withWidth` timestampWidth
 
-drawClientLine :: Double -> Double -> Double -> Render ()
-drawClientLine x y1 y2 = saved $ do
+drawClientLines :: NonEmpty Double -> Double -> Double -> Render ()
+drawClientLines xs y1 y2 = saved $ do
     setSourceRGB 0.7 0.7 0.7
-    moveTo x y1
-    lineTo x y2
-    stroke
+    forM_ (nonEmptyToList xs) $ \x -> do
+        moveTo x y1
+        lineTo x y2
+        stroke
 
 drawRule :: Double -> Double -> Double -> Render ()
 drawRule x1 x2 y = saved $ do
@@ -450,4 +511,10 @@ drawRule x1 x2 y = saved $ do
     lineTo x2 y
     stroke
 
+drawHighlight :: Rect -> Render ()
+drawHighlight (x1, y1, x2, y2) = saved $ do
+    setSourceRGB 0.8 0.9 1.0
+    rectangle x1 y1 (x2 - x1) (y2 - y1)
+    fill
+
 -- vim: sw=2 sts=2
