@@ -26,23 +26,26 @@ where
 import Control.Applicative ((<$>))
 import Control.Monad (forM_)
 import Text.Printf
-import Graphics.UI.Gtk
-import Graphics.UI.Gtk.Glade
+import Graphics.UI.Gtk hiding (Markup)
 import Bustle.Stats
 import Bustle.Types (Log)
+import qualified Bustle.Markup as Markup
+import Bustle.Markup (Markup)
+import Data.Monoid
 
 data StatsPane =
     StatsPane { spCountStore :: ListStore FrequencyInfo
               , spTimeStore :: ListStore TimeInfo
+              , spSizeStore :: ListStore SizeInfo
               }
 
-statsPaneNew :: GladeXML
+statsPaneNew :: Builder
              -> Maybe Pixbuf
              -> Maybe Pixbuf
              -> IO StatsPane
-statsPaneNew xml methodIcon signalIcon = do
-  [frequencySW, durationSW] <- mapM (xmlGetWidget xml castToScrolledWindow)
-      ["frequencySW", "durationSW"]
+statsPaneNew builder methodIcon signalIcon = do
+  [frequencySW, durationSW, sizeSW] <- mapM (builderGetObject builder castToScrolledWindow)
+      ["frequencySW", "durationSW", "sizeSW"]
 
   (countStore, countView) <- newCountView methodIcon signalIcon
   containerAdd frequencySW countView
@@ -50,7 +53,14 @@ statsPaneNew xml methodIcon signalIcon = do
   (timeStore, timeView) <- newTimeView
   containerAdd durationSW timeView
 
-  return $ StatsPane countStore timeStore
+  (sizeStore, sizeView) <- newSizeView methodIcon signalIcon
+  containerAdd sizeSW sizeView
+
+  widgetShow countView
+  widgetShow timeView
+  widgetShow sizeView
+
+  return $ StatsPane countStore timeStore sizeStore
 
 statsPaneSetMessages :: StatsPane
                      -> Log -- ^ session bus messages
@@ -61,26 +71,31 @@ statsPaneSetMessages sp sessionMessages systemMessages = do
     -- but I think that's okay for now.
     let allMessages = sessionMessages ++ systemMessages
 
+    listStoreClear (spCountStore sp)
+    listStoreClear (spTimeStore sp)
+    listStoreClear (spSizeStore sp)
+
     forM_ (frequencies allMessages) $ listStoreAppend (spCountStore sp)
     forM_ (methodTimes allMessages) $ listStoreAppend (spTimeStore sp)
+    forM_ (messageSizes allMessages) $ listStoreAppend (spSizeStore sp)
 
 addTextRenderer :: TreeViewColumn
                 -> ListStore a
                 -> Bool
-                -> (a -> String)
+                -> (a -> Markup)
                 -> IO CellRendererText
 addTextRenderer col store expand f = do
     renderer <- cellRendererTextNew
     cellLayoutPackStart col renderer expand
     set renderer [ cellTextSizePoints := 7 ]
     cellLayoutSetAttributes col renderer store $ \x ->
-        [ cellTextMarkup := Just $ f x ]
+        [ cellTextMarkup := Just . Markup.unMarkup $ f x ]
     return renderer
 
 addMemberRenderer :: TreeViewColumn
                   -> ListStore a
                   -> Bool
-                  -> (a -> String)
+                  -> (a -> Markup)
                   -> IO CellRendererText
 addMemberRenderer col store expand f = do
     renderer <- addTextRenderer col store expand f
@@ -94,7 +109,7 @@ addMemberRenderer col store expand f = do
 addStatColumn :: TreeView
               -> ListStore a
               -> String
-              -> (a -> String)
+              -> (a -> Markup)
               -> IO ()
 addStatColumn view store title f = do
     col <- treeViewColumnNew
@@ -104,6 +119,29 @@ addStatColumn view store title f = do
     treeViewAppendColumn view col
     return ()
 
+addTextStatColumn :: TreeView
+                  -> ListStore a
+                  -> String
+                  -> (a -> String)
+                  -> IO ()
+addTextStatColumn view store title f =
+    addStatColumn view store title (Markup.escape . f)
+
+-- If we managed to load the method and signal icons...
+maybeAddTypeIconColumn :: CellLayoutClass layout
+                       => layout
+                       -> ListStore a
+                       -> Maybe Pixbuf
+                       -> Maybe Pixbuf
+                       -> (a -> Bool)
+                       -> IO ()
+maybeAddTypeIconColumn nameColumn store (Just m) (Just s) isMethod = do
+    typeRenderer <- cellRendererPixbufNew
+    cellLayoutPackStart nameColumn typeRenderer False
+    cellLayoutSetAttributes nameColumn typeRenderer store $ \row ->
+            [ cellPixbuf := if isMethod row then m else s ]
+maybeAddTypeIconColumn _ _ _ _ _ = return ()
+
 newCountView :: Maybe Pixbuf
              -> Maybe Pixbuf
              -> IO (ListStore FrequencyInfo, TreeView)
@@ -119,21 +157,13 @@ newCountView method signal = do
                  , treeViewColumnExpand := True
                  ]
 
-  -- If we managed to load the method and signal icons...
-  case (method, signal) of
-      (Just m, Just s) -> do
-          typeRenderer <- cellRendererPixbufNew
-          cellLayoutPackStart nameColumn typeRenderer False
-          cellLayoutSetAttributes nameColumn typeRenderer countStore $
-              \fi ->
-                  [ cellPixbuf := case fiType fi of
-                                      TallyMethod -> m
-                                      TallySignal -> s
-                  ]
-      _ -> return ()
+  maybeAddTypeIconColumn nameColumn countStore method signal $ \fi ->
+      case fiType fi of
+          TallyMethod -> True
+          TallySignal -> False
 
   addMemberRenderer nameColumn countStore True $ \fi ->
-      fiInterface fi ++ ".<b>" ++ fiMember fi ++ "</b>"
+      Markup.formatMember (fiInterface fi) (fiMember fi)
   treeViewAppendColumn countView nameColumn
 
   countColumn <- treeViewColumnNew
@@ -172,13 +202,64 @@ newTimeView = do
                  ]
 
   addMemberRenderer nameColumn timeStore True $ \ti ->
-      tiInterface ti ++ "<b>" ++ tiMethodName ti ++ "</b>"
+      Markup.formatMember (tiInterface ti) (tiMethodName ti)
   treeViewAppendColumn timeView nameColumn
 
-  addStatColumn timeView timeStore "Total"
+  addTextStatColumn timeView timeStore "Total"
                 (printf "%.1f ms" . tiTotalTime)
-  addStatColumn timeView timeStore "Calls" (show . tiCallFrequency)
-  addStatColumn timeView timeStore "Mean"
+  addTextStatColumn timeView timeStore "Calls" (show . tiCallFrequency)
+  addTextStatColumn timeView timeStore "Mean"
                 (printf "%.1f ms" . tiMeanCallTime)
 
   return (timeStore, timeView)
+
+formatSizeInfoMember :: SizeInfo -> Markup
+formatSizeInfoMember si =
+    f (Markup.formatMember (siInterface si) (siName si))
+  where
+    f = case siType si of
+            SizeReturn -> Markup.i
+            SizeError  -> Markup.red
+            _          -> id
+
+formatSize :: Int -> Markup
+formatSize s
+    | s < maxB = value 1 `mappend` units "B"
+    | s < maxKB = value 1024 `mappend` units "KB"
+    | otherwise = value (1024 * 1024) `mappend` units "MB"
+  where
+    maxB = 10000
+    maxKB = 10000 * 1024
+
+    units = Markup.escape . (' ':)
+
+    value factor = Markup.escape (show (s `div` factor))
+
+newSizeView :: Maybe Pixbuf
+            -> Maybe Pixbuf
+            -> IO (ListStore SizeInfo, TreeView)
+newSizeView methodIcon_ signalIcon_ = do
+  sizeStore <- listStoreNew []
+  sizeView <- treeViewNewWithModel sizeStore
+
+  set sizeView [ treeViewHeadersVisible := True ]
+
+  nameColumn <- treeViewColumnNew
+  treeViewColumnSetTitle nameColumn "Member"
+  set nameColumn [ treeViewColumnResizable := True
+                 , treeViewColumnExpand := True
+                 ]
+
+  maybeAddTypeIconColumn nameColumn sizeStore methodIcon_ signalIcon_ $ \si ->
+      case siType si of
+          SizeSignal -> False
+          -- We distinguish between call, return and error by <i> and red.
+          _          -> True
+  addMemberRenderer nameColumn sizeStore True formatSizeInfoMember
+  treeViewAppendColumn sizeView nameColumn
+
+  addStatColumn sizeView sizeStore "Smallest" (formatSize . siMinSize)
+  addStatColumn sizeView sizeStore "Mean" (formatSize . siMeanSize)
+  addStatColumn sizeView sizeStore "Largest" (formatSize . siMaxSize)
+
+  return (sizeStore, sizeView)
